@@ -20,13 +20,10 @@
 
 #include <nmofono/manager-impl.h>
 #include <nmofono/connectivity-service-settings.h>
+#include <nmofono/ethernet/ethernet-link.h>
 #include <nmofono/wifi/wifi-link-impl.h>
 #include <nmofono/wwan/sim-manager.h>
-#include <NetworkManagerActiveConnectionInterface.h>
-#include <NetworkManagerDeviceInterface.h>
 #include <NetworkManagerInterface.h>
-#include <NetworkManagerSettingsInterface.h>
-#include <NetworkManagerSettingsConnectionInterface.h>
 
 #define slots
 #include <qofono-qt5/qofonomanager.h>
@@ -48,6 +45,8 @@
 #include <algorithm>
 #include <iostream>
 
+#include "nm-device-statistics-monitor.h"
+
 using namespace std;
 
 namespace nmofono {
@@ -59,7 +58,9 @@ public:
     Manager& p;
 
     shared_ptr<OrgFreedesktopNetworkManagerInterface> nm;
+    shared_ptr<OrgFreedesktopNetworkManagerSettingsInterface> m_settingsInterface;
     shared_ptr<QOfonoManager> m_ofono;
+    connection::ActiveConnectionManager::SPtr m_activeConnectionManager;
 
     bool m_flightMode = true;
     bool m_unstoppableOperationHappening = false;
@@ -72,7 +73,9 @@ public:
 
     bool m_modemAvailable = false;
 
-    QSet<Link::Ptr> m_nmLinks;
+    QList<QDBusObjectPath> m_nmDevices;
+
+    QSet<Link::SPtr> m_nmLinks;
     QMap<QString, wwan::Modem::Ptr> m_ofonoLinks;
 
     SimUnlockDialog::Ptr m_unlockDialog;
@@ -95,6 +98,8 @@ public:
 
     QTimer m_checkSimForMobileDataTimer;
 
+    NMDeviceStatisticsMonitor::Ptr m_statisticsMonitor;
+
     Private(Manager& parent) :
         p(parent)
     {
@@ -401,6 +406,19 @@ public Q_SLOTS:
             m_ofonoLinks[path] = modem;
             connect(modem.get(), &wwan::Modem::readyToUnlock, this, &Private::modemReadyToUnlock);
             connect(modem.get(), &wwan::Modem::ready, this, &Private::modemReady);
+
+            for (const auto &nmobjpath : m_nmDevices)
+            {
+                auto dev = make_shared<OrgFreedesktopNetworkManagerDeviceInterface>(
+                            NM_DBUS_SERVICE,
+                            nmobjpath.path(),
+                            nm->connection());
+                if (dev->udi() == path) {
+                    modem->setNmPath(nmobjpath.path());
+                    m_statisticsMonitor->addLink(modem);
+                    break;
+                }
+            }
         }
 
         Q_EMIT p.linksUpdated();
@@ -508,14 +526,24 @@ ManagerImpl::updateNetworkingStatus(uint status)
 ManagerImpl::ManagerImpl(notify::NotificationManager::SPtr notificationManager,
                          KillSwitch::Ptr killSwitch,
                          HotspotManager::SPtr hotspotManager,
+                         connection::ActiveConnectionManager::SPtr activeConnectionManager,
                          const QDBusConnection& systemConnection) :
         d(new ManagerImpl::Private(*this))
 {
     d->nm = make_shared<OrgFreedesktopNetworkManagerInterface>(NM_DBUS_SERVICE, NM_DBUS_PATH, systemConnection);
+    d->m_settingsInterface = make_shared<OrgFreedesktopNetworkManagerSettingsInterface>(
+                    NM_DBUS_SERVICE, NM_DBUS_PATH_SETTINGS, systemConnection);
+    d->m_activeConnectionManager = activeConnectionManager;
 
     d->m_unlockDialog = make_shared<SimUnlockDialog>(notificationManager);
     connect(d->m_unlockDialog.get(), &SimUnlockDialog::ready, d.get(), &Private::sim_unlock_ready);
 
+    d->m_statisticsMonitor = make_shared<NMDeviceStatisticsMonitor>();
+
+    connect(d->m_statisticsMonitor.get(), &NMDeviceStatisticsMonitor::txChanged, this, &ManagerImpl::txChanged);
+    connect(d->m_statisticsMonitor.get(), &NMDeviceStatisticsMonitor::rxChanged, this, &ManagerImpl::rxChanged);
+
+
     d->m_ofono = make_shared<QOfonoManager>();
 
     // Load the SIM manager before we connect to the signals
@@ -545,7 +573,7 @@ ManagerImpl::ManagerImpl(notify::NotificationManager::SPtr notificationManager,
     connect(d->m_hotspotManager.get(), &HotspotManager::reportError, this, &Manager::reportError);
 
     connect(d->nm.get(), &OrgFreedesktopNetworkManagerInterface::DeviceAdded, this, &ManagerImpl::device_added);
-    QList<QDBusObjectPath> devices(d->nm->GetDevices());
+    QList<QDBusObjectPath> devices = d->nm->GetDevices();
     for(const auto &path : devices) {
         device_added(path);
     }
@@ -570,7 +598,7 @@ ManagerImpl::ManagerImpl(notify::NotificationManager::SPtr notificationManager,
     for(auto sim : d->m_sims) {
         connect(sim.get(), &wwan::Sim::presentChanged, d.get(), &Private::startCheckSimForMobileDataTimer);
     }
-    d->m_checkSimForMobileDataTimer.start();
+    d->m_checkSimForMobileDataTimer.start();    
 }
 
 bool
@@ -690,7 +718,12 @@ void
 ManagerImpl::device_removed(const QDBusObjectPath &path)
 {
     qDebug() << "Device Removed:" << path.path();
-    Link::Ptr toRemove;
+
+    d->m_nmDevices.removeAll(path);
+
+    d->m_statisticsMonitor->remove(path.path());
+
+    Link::SPtr toRemove;
     for (auto dev : d->m_nmLinks)
     {
         auto wifiLink = dynamic_pointer_cast<wifi::WifiLinkImpl>(dev);
@@ -712,6 +745,9 @@ void
 ManagerImpl::device_added(const QDBusObjectPath &path)
 {
     qDebug() << "Device Added:" << path.path();
+
+    d->m_nmDevices.append(path);
+
     for (const auto &dev : d->m_nmLinks)
     {
         auto wifiLink = dynamic_pointer_cast<wifi::WifiLinkImpl>(dev);
@@ -721,24 +757,54 @@ ManagerImpl::device_added(const QDBusObjectPath &path)
         }
     }
 
-    Link::Ptr link;
+    Link::SPtr link;
     try {
         auto dev = make_shared<OrgFreedesktopNetworkManagerDeviceInterface>(
             NM_DBUS_SERVICE, path.path(), d->nm->connection());
-        if (dev->deviceType() == NM_DEVICE_TYPE_WIFI) {
-            wifi::WifiLink::Ptr tmp = make_shared<wifi::WifiLinkImpl>(dev,
-                                                d->nm,
-                                                d->m_killSwitch);
-
-            // We're not interested in showing access points
-            if (tmp->name() != d->m_hotspotManager->interface())
+        switch (dev->deviceType())
+        {
+            case NM_DEVICE_TYPE_WIFI:
             {
-                tmp->setDisconnectWifi(d->m_hotspotManager->disconnectWifi());
-                QObject::connect(d->m_hotspotManager.get(), &HotspotManager::disconnectWifiChanged,
-                        tmp.get(), &wifi::WifiLink::setDisconnectWifi);
+                wifi::WifiLink::SPtr tmp = make_shared<wifi::WifiLinkImpl>(dev,
+                                                    d->nm,
+                                                    d->m_killSwitch,
+                                                    d->m_activeConnectionManager);
 
-                link = tmp;
+                // We're not interested in showing access points
+                if (tmp->name() != d->m_hotspotManager->interface())
+                {
+                    tmp->setDisconnectWifi(d->m_hotspotManager->disconnectWifi());
+                    QObject::connect(d->m_hotspotManager.get(), &HotspotManager::disconnectWifiChanged,
+                            tmp.get(), &wifi::WifiLink::setDisconnectWifi);
+
+                    link = tmp;
+                }
+                break;
             }
+            case NM_DEVICE_TYPE_ETHERNET:
+            {
+                link = make_shared<ethernet::EthernetLink>(
+                        dev, d->m_settingsInterface, d->m_activeConnectionManager);
+                break;
+            }
+            case NM_DEVICE_TYPE_MODEM:
+            {
+                if (dev->driver() == "ofono")
+                {
+                    for (const auto &modem : d->m_ofonoLinks)
+                    {
+                        if (modem->ofonoPath() == dev->udi())
+                        {
+                            modem->setNmPath(path.path());
+                            d->m_statisticsMonitor->addLink(modem);
+                            break;
+                        }
+                    }
+                }
+                break;
+            }
+            default:
+                break;
         }
     } catch (const exception &e) {
         qDebug() << ": failed to create Device proxy for "<< path.path() << ": ";
@@ -750,6 +816,8 @@ ManagerImpl::device_added(const QDBusObjectPath &path)
     if (link) {
         d->m_nmLinks.insert(link);
         Q_EMIT linksUpdated();
+
+        d->m_statisticsMonitor->addLink(link);
     }
 
     d->updateHasWifi();
@@ -762,10 +830,10 @@ ManagerImpl::unstoppableOperationHappening() const
     return d->m_unstoppableOperationHappening;
 }
 
-QSet<Link::Ptr>
+QSet<Link::SPtr>
 ManagerImpl::links() const
 {
-    QSet<Link::Ptr> result(d->m_nmLinks);
+    QSet<Link::SPtr> result(d->m_nmLinks);
     for(auto i: d->m_ofonoLinks)
     {
         result.insert(i);
@@ -866,10 +934,10 @@ ManagerImpl::unlockModemByName(const QString &name)
 }
 
 
-QSet<wifi::WifiLink::Ptr>
+QSet<wifi::WifiLink::SPtr>
 ManagerImpl::wifiLinks() const
 {
-    QSet<wifi::WifiLink::Ptr> result;
+    QSet<wifi::WifiLink::SPtr> result;
     for(auto link: d->m_nmLinks)
     {
         if (link->type() == Link::Type::wifi)
@@ -880,6 +948,20 @@ ManagerImpl::wifiLinks() const
     return result;
 }
 
+QSet<ethernet::EthernetLink::SPtr>
+ManagerImpl::ethernetLinks() const
+{
+    QSet<ethernet::EthernetLink::SPtr> result;
+    for(auto link: d->m_nmLinks)
+    {
+        if (link->type() == Link::Type::ethernet)
+        {
+            result.insert(dynamic_pointer_cast<ethernet::EthernetLink>(link));
+        }
+    }
+    return result;
+}
+
 QSet<wwan::Modem::Ptr>
 ManagerImpl::modemLinks() const
 {
@@ -976,6 +1058,18 @@ ManagerImpl::sims() const
     return d->m_sims;
 }
 
+bool
+ManagerImpl::tx() const
+{
+    return d->m_statisticsMonitor->tx();
+}
+
+bool
+ManagerImpl::rx() const
+{
+    return d->m_statisticsMonitor->rx();
+}
+
 
 }
 
